# coding: utf-8

"""GirlFriend自带的Workflow实现
"""

import sys
import types
from girlfriend.exception import (
    GirlFriendBizException,
    InvalidArgumentException,
    UnknownWitchToExecuteException
)
from girlfriend.util.lang import (
    DelegateMeta,
    SequenceCollectionType,
    parse_context_var,
)
from girlfriend.util.logger import (
    create_logger,
    stdout_handler,
)
from girlfriend.workflow.protocol import (
    AbstractContext,
    AbstractJob,
    AbstractDecision,
    AbstractWorkflow,
    OkEnd,
    ErrorEnd,
    BadRequestEnd,
    AbstractListener
)
from girlfriend.plugin import plugin_manager


class Context(AbstractContext):

    """基于本地内存的上下文实现
    """

    __metaclass__ = DelegateMeta

    # 以下方法将代理到字典类
    delegate_methods = [
        "__contains__",
        "__delitem__",
        "__len__",
        "__setitem__",
        "get",
        "__iter__",
    ]

    def __init__(self, config, args, plugin_mgr, logger):
        super(Context, self).__init__()
        self.delegate = {}
        self._config = config
        self._current_unit = None
        if args is None:
            self._args = {}
        else:
            self._args = args
        self._plugin_mgr = plugin_mgr
        self._logger = logger

    def __getitem__(self, property):
        return self.delegate.get(property, None)

    @property
    def config(self):
        """配置对象
        """
        return self._config

    def logger(self):
        """获取Logger对象
        """
        return self._logger

    @property
    def current_unit(self):
        """当前正在进行的工作单元名称
        """
        return self._current_unit

    @current_unit.setter
    def current_unit(self, current_unit):
        self._current_unit = current_unit

    @property
    def current_unittype(self):
        """当前正在进行的工作单元类型
        """
        return self._current_unittype

    @current_unittype.setter
    def current_unittype(self, current_unittype):
        self._current_unittype = current_unittype

    def args(self, job_name):
        """获取某个job运行所需要的参数
        """
        return self._args.get(job_name, None)

    def plugin(self, plugin_name):
        return self._plugin_mgr.plugin(plugin_name)

    def __str__(self):
        return "<Context args={args}, config={config}, ctx={ctx}>".format(
            args=self._args,
            config=self._config,
            ctx=self.delegate
        )


class Job(AbstractJob):

    """最基本的任务单元实现
    """

    def __init__(self, name, plugin=None, caller=None, args=None, goto=None):
        """
          :param name   任务名称，在整个工作流中唯一供跳转声明使用
          :param plugin 使用的插件名称，如果不指定，在没有caller的情况下，任务名称将作为插件名称
          :param caller 可执行逻辑，在不使用插件的时候可以指定一个函数来作为任务逻辑
                        如果同时指定了caller和plugin，那么会抛出UnknowWitchToExecute异常
          :param args   执行插件所需要的参数，可以不指定，在运行时再具体指定
          :param goto   执行完毕后的下一步任务名，可以不指定，自动取任务链上下一个位置
        """

        self._name = name

        # 两者都有指定，抛出异常
        if plugin and caller:
            raise UnknownWitchToExecuteException(
                u"Job '{}' 不能同时指定caller和plugin".format(self._name)
            )

        # 两者都未指定，那么插件名称默认为Job的名称
        if not plugin and not caller:
            self._plugin_name = name
        else:
            self._plugin_name = plugin

        self._caller = caller
        self._args = args
        self._goto = goto

    @property
    def name(self):
        return self._name

    @property
    def plugin_name(self):
        return self._plugin_name

    @property
    def goto(self):
        return self._goto

    @goto.setter
    def goto(self, goto):
        self._goto = goto

    def execute(self, context):

        # 将函数类型参数展开
        if isinstance(self._args, types.FunctionType):
            self._args = self._args(context)

        # 如果是字符串类型，那么以上下文中的属性作为参数列表
        if isinstance(self._args, types.StringTypes):
            if self._args.startswith("$"):
                self._args = context[self._args[1:]]
            else:
                self._args = context[self._args]

        # 如果是生成器，那么迭代执行任务
        if isinstance(self._args, types.GeneratorType):
            return [self._execute(context, template_args)
                    for template_args in self._args]
        else:
            return self._execute(context, self._args)

    def _execute(self, context, template_args):
        args = self._get_runtime_args(context, template_args)

        # 获取可执行对象
        executable = self._get_executable(context)

        result = None
        if args is None:
            result = executable(context)
        elif isinstance(args, SequenceCollectionType):
            result = executable(context, *args)
        elif isinstance(args, types.DictType):
            result = executable(context, **args)

        # 自动将插件计算结果写入Context
        context["{}.result".format(self.name)] = result
        return result

    def _get_executable(self, context):
        if self._caller:
            return self._caller
        plugin = context.plugin(self._plugin_name)
        return plugin.execute

    def _get_runtime_args(self, context, template_args):
        """获取运行时参数
           Job的参数分为两部分，一部分是在workflow声明时指定的参数
           一部分是在运行时通过context指定的参数

           参数支持两种形式，一种是参数列表，一种是字典关键字的形式

           如果是列表，那么将会用context中的列表取代self._args中的列表
           如果是字典，那么将会执行update操作，context中的项将覆盖_args中的项
           如果self._args为None，那么直接使用context中的参数
           如果self._args与context类型不一致，那么抛出异常

           最终将args中的字符串变量名用Context真实值进行替换
        """

        context_args = context.args(self.name)
        if isinstance(context_args, types.FunctionType):
            context_args = context_args(context)

        if isinstance(context_args, types.StringTypes):
            if context_args.startswith("$"):
                context_args = context[context_args[1:]]
            else:
                context_args = context[context_args]

        args_schema = None
        if not context_args:
            args_schema = template_args
        elif not template_args:
            args_schema = context_args
        elif isinstance(template_args, SequenceCollectionType):
            if not isinstance(context_args, SequenceCollectionType):
                raise InvalidArgumentException(
                    u"Job '{}'的初始参数类型跟上下文指定的参数类型不一致".format(self.name)
                )
            args_schema = template_args
            if context_args:
                args_schema = context_args
        elif isinstance(template_args, types.DictType):
            if not isinstance(context_args, types.DictType):
                raise InvalidArgumentException(
                    u"Job '{}'的初始参数类型跟上下文指定的参数类型不一致".format(self.name)
                )
            args_schema = template_args
            if context_args:
                args_schema = dict(template_args)
                args_schema.update(context_args)
        else:
            raise InvalidArgumentException(u"只能接受列表、元组、字典类型的参数对象")

        # 替换为真正的变量
        if not args_schema:
            return None
        elif isinstance(args_schema, SequenceCollectionType):
            return [parse_context_var(context, arg) for arg in args_schema]
        elif isinstance(args_schema, types.DictType):
            return {arg_name: parse_context_var(context, args_schema[arg_name])
                    for arg_name in args_schema}


class Decision(AbstractDecision):

    """判断节点的实现
    """

    def __init__(self, name, decide_logic):
        """
            :param name 节点名称
            :param decide_logic 判断逻辑，接受一个context参数，
                                对当前工作流的状态进行判断并返回下一步要执行的节点名称
        """
        self._name = name
        self._decide_logic = decide_logic

    @property
    def name(self):
        return self._name

    def execute(self, context):
        return self._decide_logic(context)


class Workflow(AbstractWorkflow):

    """无状态的本地工作流执行引擎实现
    """

    def __init__(self, workflow_list, config,
                 plugin_mgr=plugin_manager, context_factory=Context,
                 logger=None):
        """
        :param workflow_list 工作单元列表
        :param config 配置数据
        :param plugin_mgr 插件管理器，默认是plugin自带的entry_points管理器
        :param context_factory 上下文工厂，必须具有config, args, plugin_mgr, parent
                               这四个约定的参数
        :param logger 日志对象
        """

        self._workflow_list = workflow_list
        self._config = config
        self._plugin_manager = plugin_mgr

        self._units = {}  # 以名称为Key的工作流单元引用字典
        for idx, unit in enumerate(self._workflow_list):
            if unit.name in self._units:
                raise WorkflowUnitExistedException(unit.name)
            self._units[unit.name] = unit
            if unit.unittype == "job":
                # 如果未指定goto，那么goto的默认值是下一个节点
                if unit.goto is None:
                    if idx < len(self._workflow_list) - 1:
                        unit.goto = self._workflow_list[idx + 1].name
                    else:
                        unit.goto = "end"
        self._context_factory = context_factory
        self._listeners = []

        # 创建logger
        if logger is None:
            self._logger = create_logger("girlfriend", (stdout_handler(),))
        else:
            self._logger = logger

    def add_listener(self, listener=None, **kws):
        """为工作流添加监听器，该方法有两种使用方式。

           直接使用: 直接传递一个监听器对象或者一个AbstractListener子类的类型对象
                    当传递监听器对象时，此对象会在每次进行execute时重复使用。
                    当传递类型对象时，每次进行execute都会依照此类型生成一个新的对象
                    类型对象的构造函数参数除self以外必须为空。
           包装函数: 只传递感兴趣的生命周期
                    workflow.add_listener(start=on_start, finish=on_finish)
        """
        if listener:
            if (not isinstance(listener, AbstractListener) and
                    not issubclass(listener, AbstractListener)):
                raise InvalidArgumentException(
                    u"监听器必须是AbstractListener对象或者是其子类类型对象")
            self._listeners.append(listener)
            return
        if kws:
            event_funcs = []
            for event_name in kws:
                if not event_name.startswith("on_"):
                    event_name = "on_" + event_name
                if event_name not in AbstractListener.EVENTS:
                    raise InvalidArgumentException(u"不被支持的事件:" + event_name)
                event_funcs += (event_name, kws[event_name])
            self._listeners.append(AbstractListener.wrap_function(event_funcs))
            return

    def execute(self, args=None, start_point=None):
        """执行工作流
           :param start_point 工作流起始点，如果为None，
                              那么将从workflow_list第一个元素开始执行
           :param args 运行时参数
        """
        if start_point:
            start_unit = self._units[start_point]
        else:
            start_unit = self._workflow_list[0]

        # 构建上下文
        ctx = self._context_factory(
            config=self._config,
            args=args,
            plugin_mgr=self._plugin_manager,
            logger=self._logger
        )
        listener_objects = {}  # 用来保存每次执行时需要创建新对象的listener

        self._logger.info(u"工作流开始执行，起始点为 '{}'".format(start_unit.name))

        # 执行初始化的listener
        self._execute_listeners("on_start", ctx, listener_objects)

        current_unit = start_unit
        last_result = None
        goto = None
        while True:
            ctx.current_unit = current_unit.name
            ctx.current_unittype = current_unit.unittype

            # 进入新的Unit
            self._logger.info(u"开始执行工作单元 {} [{}]".format(
                current_unit.name, current_unit.unittype))

            self._execute_listeners("on_unit_start", ctx, listener_objects)

            try:
                if current_unit.unittype in ("job", "waiting_for"):
                    last_result = current_unit.execute(ctx)
                    goto = current_unit.goto
                elif current_unit.unittype == "decision":
                    goto = current_unit.execute(ctx)
                elif current_unit.unittype == "fork":
                    pass
                elif current_unit.unittype == "end":
                    # 用户指定的结束节点
                    current_unit.execute(ctx)
                    self._execute_listeners("on_unit_finish",
                                            ctx, listener_objects)
                    self._execute_listeners("on_finish", ctx, listener_objects)
                    self._logger.info(u"工作流顺利执行完毕")
                    return current_unit
            except InvalidArgumentException as e:
                self._logger.exception(u"单元参数错误")
                exc_type, exc_value, tb = sys.exc_info()
                self._execute_on_error_listeners(ctx,
                                                 exc_type, exc_value, tb,
                                                 listener_objects)
                return BadRequestEnd(msg=unicode(e))
            except Exception:
                self._logger.exception(u"系统错误，工作流被迫中止")
                exc_type, exc_value, tb = sys.exc_info()
                self._execute_on_error_listeners(ctx,
                                                 exc_type, exc_value, tb,
                                                 listener_objects)
                return ErrorEnd(exc_type, exc_value, tb)  # 返回错误的结果
            else:
                # 执行完成事件
                self._execute_listeners("on_unit_finish",
                                        ctx, listener_objects)
                self._logger.info(u"工作单元 {} [{}] 执行完毕".format(
                    current_unit.name, current_unit.unittype))
                if goto == "end":
                    self._execute_listeners("on_finish", ctx, listener_objects)
                    self._logger.info(u"工作流顺利执行完毕")
                    return OkEnd(result=last_result)  # 将最后一个单元的结果作为默认返回的结果
                else:
                    current_unit = self._units[goto]  # 处理下一个单元

    def _execute_listeners(self, event_name, context, listener_objects):
        """遍历执行监听器
        """
        for idx, listener in enumerate(self._listeners):
            if isinstance(listener, AbstractListener):
                getattr(listener, event_name)(context)
            elif issubclass(listener, AbstractListener):
                listener_obj = self.__create_listener_obj(
                    idx, listener, listener_objects)
                getattr(listener_obj, event_name)(context)

    def _execute_on_error_listeners(self, context,
                                    exc_type, exc_value, tb, listener_objects):
        """遍历执行错误监听器
        """
        for idx, listener in enumerate(self._listeners):
            if isinstance(listener, AbstractListener):
                listener.on_error(context, exc_type, exc_value, tb)
            elif issubclass(listener, AbstractListener):
                listener_obj = self.__create_listener_obj(
                    idx, listener, listener_objects)
                listener_obj.on_error(
                    context, exc_type, exc_value, tb)

    def __create_listener_obj(self, idx, listener_class, listener_objects):
        listener_obj = listener_objects.get(idx)
        if not listener_obj:
            listener_obj = listener_class()
            listener_objects[idx] = listener_obj
        return listener_obj


class WorkflowUnitExistedException(GirlFriendBizException):

    def __init__(self, unit_name):
        super(WorkflowUnitExistedException, self).__init__(
            u"工作单元 '{}' 已经存在".format(unit_name))



# coding: utf-8

"""使用此工具，可以直接通过命令行来执行一个工作流

   gf_workflow -m workflow.py -c config_file -e test ...
"""

import imp
import sys
import time
import types
import daemon
import os.path
import logging
import argparse
import pkg_resources
from girlfriend.util.script import show_msg_and_exit
from girlfriend.util.config import Config
from girlfriend.util.cmdargs import print_help
from girlfriend.util.time import parse_time_unit
from girlfriend.util.logger import (
    create_logger,
    stdout_handler,
    daily_rotaiting_handler,
    get_logger_level_by_name,
)
from girlfriend.workflow.protocol import Env
from girlfriend.workflow.gfworkflow import Workflow, Context
from girlfriend.plugin import plugin_manager as DEFAULT_PLUGIN_MANAGER


# 插件管理器以及用到的插件名称
plugin_manager, plugin_names = None, None


def main():
    tools_options = _parse_arguments()
    config = Config.load_by_name(tools_options.config)
    _add_python_path(tools_options, config)
    workflow_module = _load_module(tools_options)

    workflow_parser = getattr(workflow_module, "cmd_parser", None)

    # 展示工作流模块所需要的参数说明
    if tools_options.show_args:
        _show_args(workflow_parser)

    # 解析workflow所需参数
    workflow_options = None
    if workflow_parser is not None:
        workflow_options = workflow_parser.parse_known_args(sys.argv[1:])[0]

    # 获取当前运行环境
    current_env = _get_current_env(workflow_module, tools_options)

    # 更新配置信息，使用模块的配置项覆盖配置文件中的配置项
    _update_config(config, current_env, workflow_module, workflow_options)

    # 以Daemon形式运行
    if tools_options.daemon:
        with daemon.DaemonContext():
            _execute_workflow(config, workflow_module, workflow_options,
                              tools_options, current_env)
    else:
        _execute_workflow(config, workflow_module, workflow_options,
                          tools_options, current_env)


def _parse_arguments():
    """解析gf_workflow自身参数"""
    parser = argparse.ArgumentParser(
        description=__doc__.decode("utf-8"))

    parser.add_argument("--module", "-m", dest="module", help=u"包含工作流定义的模块")
    parser.add_argument("--config", "-c", dest="config",
                        help=u"配置文件，默认将使用HOME/.gf/gf.cfg", default="default")
    parser.add_argument("--environ", "-e", dest="environ", help=u"指定当前使用环境",
                        default="test")
    parser.add_argument("--path", "-p", dest="path",
                        help=u"指定PYTHONPATH，使用分号分割", default="")
    parser.add_argument("--run-mode", "-r", dest="run_mode",
                        help=(
                            u"once - 只运行一次, "
                            u"forever - 在循环中运行, "
                            u"interval:1m - 每隔一分钟运行一次"
                        ),
                        default="once")
    parser.add_argument("--show-args", dest="show_args",
                        action="store_true", help=u"显示工作流模块自定义参数")
    parser.add_argument("--daemon", "-d", dest="daemon",
                        action="store_true", help=u"以daemon的形式来运行工作流")

    return parser.parse_known_args(sys.argv[1:])[0]


def _add_python_path(tools_options, config):
    """添加PYTHONPATH"""
    python_path = tools_options.path
    if python_path:
        sys.path.extend(python_path.split(";"))
    python_path = config.get("workflow", "path")
    if python_path:
        sys.path.extend(python_path.split(";"))


def _load_module(tools_options):
    """加载工作流模块"""
    module_path = tools_options.module

    if not module_path:
        show_msg_and_exit(u"必须使用-m参数指定一个工作流描述模块")

    if module_path.startswith(":"):
        module_path = module_path[1:]
        for ep in pkg_resources.iter_entry_points("girlfriend.workflow"):
            if ep.name == module_path:
                return ep.load()
        show_msg_and_exit(u"找不到工作流模块 '{}'".format(module_path))

    # 以.py结尾，那么按照python文件的形式进行加载
    if module_path.endswith(".py"):
        if not os.path.exists(module_path):
            show_msg_and_exit(u"找不到工作流描述文件 '{}'".format(module_path))
        return imp.load_source("workflow", module_path)

    # 按照模块名称方式进行加载
    try:
        return __import__(module_path, fromlist=[""])
    except ImportError:
        show_msg_and_exit(u"找不到工作流模块 '{}'".format(module_path))


def _show_args(workflow_parser):
    """展示模块参数"""
    if workflow_parser is not None:
        print u"工作流的参数说明："
        print_help(workflow_parser)
        exit(0)
    else:
        print u"该工作流模块不需要参数"
        exit(0)


def _update_config(config, current_env, workflow_module, workflow_options):
    """更新配置信息，先按模块中的config变量进行更新，再按照具体的环境进行更新"""
    _update_config_items(config, getattr(workflow_module, "config", None),
                         workflow_options)
    _update_config_items(config, current_env.config, workflow_options)


def _update_config_items(config, new_config, workflow_options):
    if new_config is None:
        return
    if isinstance(new_config, types.FunctionType):
        new_config = new_config(workflow_options)
        if new_config is None:
            return
    config.update(new_config)


def _get_current_env(workflow_module, tools_options):
    env_list = getattr(workflow_module, "env", None)
    if env_list is None:
        return Env.test_env()
    current_env_name = tools_options.environ
    for env in env_list:
        if env.name == current_env_name:
            return env
    show_msg_and_exit(u"找不到目标环境：'{}'".format(current_env_name))


def _execute_workflow(config, workflow_module, workflow_options,
                      tools_options, current_env):
    workflow_engine = _get_workflow_engine(
        config, workflow_module, workflow_options, current_env)

    # 一次性执行
    try:
        if tools_options.run_mode == "once":
            return _run_once(workflow_engine, config, workflow_module,
                             workflow_options, current_env)
        # 永久性执行
        elif tools_options.run_mode == "forever":
            _run_forever(workflow_engine, config, workflow_module,
                         workflow_options, current_env)
        # 按时间间隔执行
        elif tools_options.run_mode.startswith("interval:"):
            _run_interval(
                workflow_engine, config,
                workflow_module, workflow_options,
                current_env, tools_options.run_mode[len("interval:"):])
        else:
            print u"未知的运行模式：'{}'".format(tools_options.run_mode)
    finally:
        _clean_plugins(config)


def _run_interval(workflow_engine, config, workflow_module,
                  workflow_options, current_env, time_unit):
    """按周期运行"""
    sleeping_seconds = parse_time_unit(time_unit)
    while True:
        _run_once(workflow_engine, config, workflow_module,
                  workflow_options, current_env)
        time.sleep(sleeping_seconds)


def _run_forever(workflow_engine, config, workflow_module,
                 workflow_options, current_env):
    """永久运行"""
    while True:
        _run_once(workflow_engine, config, workflow_module,
                  workflow_options, current_env)


def _run_once(workflow_engine, config, workflow_module,
              workflow_options, current_env):
    runtime_args = _get_runtime_args(
        config, workflow_module, workflow_options, current_env)
    return workflow_engine.execute(runtime_args)


def _get_workflow_engine(config, workflow_module,
                         workflow_options, current_env):
    """获取工作流执行引擎"""

    # 获取工作单元列表
    workflow_list = getattr(workflow_module, "workflow", None)
    if not workflow_list:
        show_msg_and_exit(u"工作流单元列表不能为空")
    if isinstance(workflow_list, types.FunctionType):
        workflow_list = workflow_list(workflow_options)
        if not workflow_list:
            show_msg_and_exit(u"工作流单元列表不能为空")

    # 获取并初始化插件管理器
    global plugin_manager
    plugin_manager = getattr(
        workflow_module, "plugin_manager", DEFAULT_PLUGIN_MANAGER)
    global plugin_names
    plugin_names = set(work_unit.plugin_name for work_unit in workflow_list
                       if work_unit.unittype == "job" and
                       work_unit.plugin_name)
    # 额外插件列表
    extra_plugin_names = getattr(workflow_module, "plugins", tuple())
    for plugin_name in extra_plugin_names:
        plugin_names.add(plugin_name)
    plugin_manager.sys_prepare(config, *plugin_names)

    # 获取logger
    logger = getattr(workflow_module, "logger", None)
    if isinstance(logger, types.FunctionType):
        logger = logger(workflow_options)
    logger_level = getattr(workflow_module, "logger_level", logging.INFO)
    if isinstance(logger_level, types.FunctionType):
        logger_level = logger_level(workflow_options)
    if isinstance(logger_level, str):
        logger_level = get_logger_level_by_name(logger_level)
    if logger is None:
        logger = create_logger("girlfriend", (stdout_handler(),),
                               level=logger_level)
    elif isinstance(logger, str):
        logger = create_logger(
            "girlfriend", (daily_rotaiting_handler(logger),),
            level=logger_level)

    workflow_engine = Workflow(
        workflow_list, config, plugin_manager, Context, logger)

    # 获取监听器列表
    listeners = getattr(workflow_module, "listeners", [])
    if isinstance(listeners, types.FunctionType):
        listeners = listeners(workflow_options)
        if not listeners:
            listeners = []

    for listener in listeners:
        workflow_engine.add_listener(listener)

    return workflow_engine


def _get_runtime_args(config, workflow_module, workflow_options, current_env):
    """先获取工作流脚本中的args属性，然后再用env中的args进行更新"""

    # 获取脚本中的args属性
    args = getattr(workflow_module, "args", {})
    if isinstance(args, types.FunctionType):
        args = args(workflow_options)
    if args is None:
        args = {}

    # 获取当前环境中的args属性
    env_args = {} if current_env.args is None else current_env.args
    if isinstance(env_args, types.FunctionType):
        env_args = env_args(workflow_options)
    if env_args is None:
        env_args = {}

    args.update(env_args)
    return args


def _clean_plugins(config):
    # 清理插件
    global plugin_manager, plugin_names
    plugin_manager.sys_cleanup(config, *plugin_names)

if __name__ == "__main__":
    main()



# coding: utf-8

"""本模块包含Table数据结构的抽象以及一个基于二维列表的Table实现，
很多以表格为核心的插件比如excel都会依赖Table结构。
"""

import types
import operator
import prettytable
from abc import (
    ABCMeta,
    abstractproperty,
    abstractmethod
)
from girlfriend.exception import InvalidTypeException
from girlfriend.data.exception import (
    IndexOutOfBoundsException,
    MissingKeyException,
    InvalidSizeException
)
from girlfriend.util.lang import SequenceCollectionType


class AbstractTable(object):

    """抽象的Table接口
       在构建新的Table结构和依赖Table的插件时,应尽量依赖于此约定展开工作
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractproperty
    def name(self):
        """表格的名称
        """
        pass

    @abstractproperty
    def titles(self):
        """以可迭代对象的形式返回表格的列名
        """
        pass

    @abstractproperty
    def row_num(self):
        """表格行数
        """
        pass

    @abstractproperty
    def column_num(self):
        """表格列数
        """
        pass

    def __len__(self):
        """表格的长度默认为行数
        """
        return self.row_num

    @abstractmethod
    def row(self, index):
        """获取某一行
        """
        pass

    @abstractmethod
    def __getitem__(self, index):
        """根据索引获取某一行
        """
        pass

    @abstractmethod
    def append(self, row):
        """添加行
        """
        pass

    @abstractmethod
    def __iter__(self):
        pass


class Title(object):

    """表格标题"""

    def __init__(self, name, title=None):
        """
        :param name: 标题名,可以方便用来进行变量引用,比如id、age、grade等等
        :param title: 供显示的标题名,比如编号/年龄/等级,可以不指定,默认为name的值
        """
        self._name = name
        if title is None:
            self._title = name
        else:
            self._title = title

    @property
    def name(self):
        return self._name

    @property
    def title(self):
        return self._title

    def __repr__(self):
        return self._title

    def __str__(self):
        return self._title


class Row(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __getattr__(self, key):
        pass

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __iter__(self, key):
        pass


class BaseLocalTable(AbstractTable):

    """基于本地对象的Table实现
    """

    __metaclass__ = ABCMeta

    def __init__(self, name, titles, data, row_type):
        self._name = name
        self._titles = titles
        if data is None:
            self._data = []
        else:
            self._data = data
        self._row_type = row_type
        self._mapping = self._gen_mapping()

    @abstractmethod
    def _gen_mapping(self):
        pass

    @property
    def name(self):
        return self._name

    @property
    def titles(self):
        return self._titles

    @property
    def row_num(self):
        return len(self._data)

    @property
    def column_num(self):
        return len(self._titles)

    def row(self, row_index):
        self._check_row_index(row_index)
        return self._row_type(self._data[row_index], self._mapping)

    def __getitem__(self, row_index):
        return self.row(row_index)

    def _check_row_index(self, row_index):
        if row_index >= self.row_num:
            raise IndexOutOfBoundsException(
                u"row_index={0}超出了边界，row_num={1}".format(
                    row_index,
                    self.row_num
                ))

    def _check_col_index(self, column_index):
        if column_index >= self.column_num:
            raise IndexOutOfBoundsException(
                u"column_index={0}，column_num={1}".format(
                    column_index,
                    self.column_num
                ))

    def __iter__(self):
        for row in self._data:
            yield self._row_type(row, self._mapping)

    def __str__(self):
        ptable = prettytable.PrettyTable(title.title for title in self.titles)
        for row in self:
            ptable.add_row(row)
        return str(ptable)


class BaseLocalRow(Row):

    __metaclass__ = ABCMeta

    def __init__(self, row_object, mapping):
        self._row = row_object
        self._mapping = mapping

    @property
    def obj(self):
        """返回被包装的对象"""
        return self._row

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __getattr__(self, attr_name):
        pass

    def __len__(self):
        return len(self._mapping)

    def __iter__(self):
        return iter(self._row)

    def __repr__(self):
        return repr(self._row)

    def __str__(self):
        return str(self._row)


class TableWrapper(object):

    """表格包装器，能够根据列表中元素的类型，自动选择对应的表格类型进行包装
    """

    def __init__(self, name, titles, table_type=None, auto_title_name=False):
        self._name = name
        self._titles = self._handle_titles(titles, auto_title_name)
        self._table_type = table_type

    def __call__(self, data):
        table_type = self._get_table_type(data)
        return table_type(
            name=self._name,
            titles=self._titles,
            data=data
        )

    def _handle_titles(self, titles, auto_title_name):
        """对Title进行处理，接受两种形式的title
           接受纯Title对象的列表或者元组，例如:
           (Title("id", u"编号"), Title("name", u"姓名"), Title("grade", u"年级"))
           也接受字符串描述的元组:
           ("id", u"编号", "name", u"姓名", "grade", u"年级")
        """
        first_ele = titles[0]
        if isinstance(first_ele, Title):
            return titles
        elif isinstance(first_ele, types.StringTypes):
            if auto_title_name:
                return [Title("field_%d" % idx, arg)
                        for idx, arg in enumerate(titles)]
            else:
                return [Title(*arg) for arg in zip(titles[::2], titles[1::2])]
        else:
            raise InvalidTypeException(u"title列表元素只允许字符串和Title类型")

    def _get_table_type(self, data):
        # 自动推断应该使用的table类型
        if self._table_type is not None:
            return self._table_type

        # 空列表
        if data is None or len(data) == 0:
            return ListTable

        # 默认取第一行进行推断
        row = data[0]
        if isinstance(row, SequenceCollectionType):
            return ListTable
        elif isinstance(row, types.DictType):
            return DictTable
        else:
            return ObjectTable


class ListTable(BaseLocalTable):

    """基于二维list的Table实现
    """

    def __init__(self, name, titles, data=None):
        """
        :param name 表格名称
        :param titles 表格标题
        :param data 如果data为None，那么会生成一个新的二维列表
                    如果不为None，那么包装既有的二维列表
        """
        BaseLocalTable.__init__(self, name, titles, data, ListRow)

    def _gen_mapping(self):
        return {title.name: idx for idx, title in enumerate(self.titles)}

    def cell(self, row_index, column_index):
        self._check_row_index(row_index)
        self._check_col_index(column_index)
        return self._data[row_index][column_index]

    def append(self, row):
        """添加新的一行
        """
        if not isinstance(row, SequenceCollectionType):
            raise InvalidTypeException(u"新行的类型必须是list或者tuple类型")
        if len(row) > self.column_num:
            raise InvalidSizeException(
                u"新行的列数为{0}，表格的列数为{1}，两者不一致".format(
                    len(row),
                    self.column_num
                ))
        self._data.append(row)


class ListRow(BaseLocalRow):

    """基于List/Tuple实现的行对象
    """

    def __init__(self, row_list, mapping):
        BaseLocalRow.__init__(self, row_list, mapping)

    def __getitem__(self, key):
        if isinstance(key, (types.IntType, types.LongType)):
            return self._row[key]
        elif isinstance(key, (types.StringType)):
            index = self._mapping.get(key, None)
            if index is None:
                raise MissingKeyException(
                    u"找不到与key={}相关联的值".format(key))
            return self._row[index]
        elif isinstance(key, SequenceCollectionType):
            index_list = (self._mapping.get(k) for k in key)
            return tuple(self._row[index] for index in index_list)
        else:
            raise InvalidTypeException(
                u"不合法的key类型：{}".format(type(key).__name__))

    def __getattr__(self, key):
        index = self._mapping.get(key, None)
        if index is None:
            raise AttributeError(u"找不到属性{}".format(key))
        return self._row[index]


class ObjectTable(BaseLocalTable):

    """ObjectTable中的每一行都是一个具体的对象
    """

    def __init__(self, name, titles, data=None):
        BaseLocalTable.__init__(self, name, titles, data, ObjectRow)

    def _gen_mapping(self):
        return [title.name for title in self.titles]

    def cell(self, row_index, column_index):
        self._check_row_index(row_index)
        self._check_col_index(column_index)
        obj = self._data[row_index]
        attr_name = self._mapping[row_index]
        return getattr(obj, attr_name)

    def append(self, row):
        self._data.append(row)


class ObjectRow(BaseLocalRow):

    def __init__(self, row_obj, mapping):
        BaseLocalRow.__init__(self, row_obj, mapping)

    def __getitem__(self, key):
        if isinstance(key, types.StringTypes):
            return getattr(self._row, key)
        if isinstance(key, (types.IntType, types.LongType)):
            key = self._mapping[key]
            return getattr(self._row, key)
        if isinstance(key, SequenceCollectionType):
            getter = operator.attrgetter(*key)
            return getter(self._row)

    def __getattr__(self, key):
        return getattr(self._row, key)

    def __iter__(self):
        for i in xrange(0, len(self._mapping)):
            yield getattr(self._row, self._mapping[i])


class DictTable(BaseLocalTable):

    """DictTable的每一行都是一个字典对象
    """

    def __init__(self, name, titles, data=None):
        BaseLocalTable.__init__(self, name, titles, data, DictRow)

    def _gen_mapping(self):
        return [title.name for title in self._titles]

    def cell(self, row_index, column_index):
        self._check_row_index(row_index)
        self._check_col_index(column_index)
        row = self._data[row_index]
        key = self._mapping(column_index)
        return row[key]

    def append(self, row):
        if not isinstance(row, types.DictType):
            raise InvalidTypeException(u"DictTable只能接受字典类型的行对象")
        if len(row) != self.column_num:
            raise InvalidSizeException(
                u"新行的列数为{0}，表格的列数为{1}，两者不一致".format(len(row), self.column_num)
            )
        self._data.append(row)


class DictRow(BaseLocalRow):

    def __init__(self, row_obj, mapping):
        BaseLocalRow.__init__(self, row_obj, mapping)

    def __getitem__(self, key):
        if isinstance(key, types.StringTypes):
            return self._row[key]
        if isinstance(key, (types.IntType, types.LongType)):
            key = self._mapping[key]
            return self._row[key]
        if isinstance(key, SequenceCollectionType):
            return tuple(self._row[k] for k in key)

    def __getattr__(self, key):
        return self._row[key]

    def __iter__(self):
        for i in xrange(len(self._mapping)):
            yield self._row[self._mapping[i]]



# coding: utf-8

"""时间操作工具
"""

import re
from girlfriend.exception import InvalidArgumentException

time_unit_regex = re.compile(r"^(\d+)(d|h|m|s)$", re.IGNORECASE)

time_units = {
    "d": 24 * 60 * 60,
    "h": 60 * 60,
    "m": 60,
    "s": 1
}


def parse_time_unit(time_unit):
    """将以下单位描述转换为秒
       1d -> 24 * 60 * 60
       1h -> 1 * 60 * 60
       1m -> 1 * 60
       1s -> 1
    """
    match_result = time_unit_regex.search(time_unit)
    if not match_result:
        raise InvalidArgumentException(u"参数 '{}' 不符合时间单位格式".format(time_unit))
    time_value, unit = (int(match_result.group(1)),
                        match_result.group(2).lower())
    return time_value * time_units[unit]



# coding: utf-8

"""邮件发送相关插件
"""

import types
import smtplib
import os.path
from abc import (
    ABCMeta,
    abstractproperty,
)
from functools import partial
from email import encoders
from email.Header import Header
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from girlfriend.util.lang import args2fields
from girlfriend.exception import InvalidArgumentException
from girlfriend.util.validating import Rule


class SMTPManager(object):
    """管理配置文件中所有登记的SMTP服务信息"""

    config_rules = (
        # SMTP 主机地址
        Rule("host", type=types.StringTypes, required=True),
        # SMTP 端口号
        Rule("port", type=(types.StringTypes, int), required=False,
             regex=r"^\d+$", default=25),
        # SMTP 账号
        Rule("account", type=types.StringTypes, required=True),
        # SMTP 密码
        Rule("password", type=types.StringTypes, required=True),
        # 是否使用SSL通道发送
        Rule("ssl", type=types.StringTypes, required=False,
             regex=r"^(true|false)$", default="false")
    )

    def __init__(self):
        self._all_smtp_config = {}
        self._validated = False  # 尚未进行验证

    def validate_config(self, config):
        """验证配置，并将配置添加到容器之中"""
        if self._validated:
            return
        for section in config.prefix("smtp_"):
            smtp_config_items = config[section]
            for rule in SMTPManager.config_rules:
                item_value = smtp_config_items.get(rule.name)
                rule.validate(item_value)
                if item_value is None:
                    smtp_config_items[rule.name] = rule.default
                if rule.name == "port" and isinstance(
                        item_value, types.StringTypes):
                    smtp_config_items["port"] = int(item_value)
            smtp_config = SMTPConfig(**smtp_config_items)
            self._all_smtp_config[section] = smtp_config
        self._validated = True

    def get_smtp_config(self, smtp_server_name):
        return self._all_smtp_config[smtp_server_name]


class SMTPConfig(object):
    """SMTP配置信息"""

    @args2fields()
    def __init__(self, host, port, account, password, ssl):
        pass
        self._ssl = (ssl.lower() == "true")

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def account(self):
        return self._account

    @property
    def password(self):
        return self._password

    @property
    def ssl(self):
        return self._ssl


_smtp_manager = SMTPManager()


class SendMailPlugin(object):

    """邮件发送插件，支持附件和模板渲染
       使用方法如下：

       Job(
          "send_mail",
          args = {
              "server": "test_smtp"
              "sender": "sam@gf.com",
              "receiver": "$users",
              "subject": subject_generator,
              "content": content_generator,
              "encoding": "utf-8",
              "attachments": (
                  Attachment(...),
                  Attachment(...),
                  Attachment(...)
              )
          }
       )

       或者：

       Job(
          "send_mail",
          args = {
              "server": "test_smtp"
              "receiver": "$users",
              "mail": TestMail
          }
       )
    """

    name = "send_mail"

    @staticmethod
    def config_validator(config):
        global _smtp_manager
        _smtp_manager.validate_config(config)

    def execute(self, context, server, receivers, mail=None,
                sender=None, subject=None, content=None,
                encoding="utf-8", attachments=None):
        """
        :param server SMTP服务名称
        :param receiver 收件人列表，可以是字符串组成的邮箱列表，也可以是对象列表
        :param mail  用来动态表述邮件内容的邮件类
        :param sender 发件人，可以传递字符串或函数，函数接受上下文以及收件人对象作为参数
        :param subject 标题，可传递字符串或函数，函数参数同上
        :param content 正文，可以是字符串或函数，函数参数同上
        :param encoding 编码，可以是字符串或函数，函数参数同上
        :param attachments 附件列表，可以是Attachment对象列表或者函数，函数参数同上
        """

        # 组装Mail对象列表
        mail_list = []
        if mail is not None:
            if not issubclass(mail, Mail):
                raise InvalidArgumentException(
                    u"mail参数必须是girlfriend.plugin.mail.Mail类型的子类")
        else:
            mail = partial(_Mail, sender=sender, subject=subject,
                           content=content, encoding=encoding,
                           attachments=attachments)

        if isinstance(receivers, types.StringTypes):
            mail_list = [mail(context=context, receiver=receivers)]
        else:
            mail_list = [mail(context=context, receiver=receiver)
                         for receiver in receivers]

        # 创建SMTP连接
        smtp_config = _smtp_manager.get_smtp_config("smtp_" + server)
        if smtp_config.ssl:
            smtp_server = smtplib.SMTP_SSL(
                host=smtp_config.host,
                port=smtp_config.port
            )
        else:
            smtp_server = smtplib.SMTP(
                host=smtp_config.host,
                port=smtp_config.port
            )
        smtp_server.login(smtp_config.account, smtp_config.password)

        try:
            for mail in mail_list:
                msg = MIMEMultipart()
                msg['From'] = mail.sender
                msg['To'] = mail.receiver_email
                msg['Subject'] = Header(mail.subject, mail.encoding)
                msg.attach(MIMEText(mail.content, "html", mail.encoding))
                attachments = mail.attachments
                if attachments:
                    for attachment in attachments:
                        msg.attach(attachment.build_mime_object())
                smtp_server.sendmail(
                    mail.sender,
                    [email_address.strip()
                     for email_address in mail.receiver_email.split(",")],
                    msg.as_string())
        finally:
            smtp_server.quit()


class Mail(object):

    __metaclass__ = ABCMeta

    def __init__(self, context, receiver):
        self._context = context
        self._receiver = receiver

    @abstractproperty
    def sender(self):
        """发件人地址"""
        pass

    @abstractproperty
    def receiver_email(self):
        """收件人邮箱地址"""
        pass

    @abstractproperty
    def subject(self):
        """邮件标题"""
        pass

    @abstractproperty
    def content(self):
        """邮件正文"""
        pass

    @property
    def attachments(self):
        """邮件附件"""
        return []

    @property
    def encoding(self):
        """编码，默认为UTF-8，如果你家Boss或PM喜欢用Mac装Windows，那么可以覆盖为gb2312"""
        return "utf-8"


class _Mail(Mail):

    """内置的Mail实现"""

    def __init__(self, context, receiver, sender, subject,
                 content, encoding, attachments):
        super(_Mail, self).__init__(context, receiver)
        self._sender = sender
        self._subject = subject
        self._content = content
        self._encoding = encoding
        self._attachments = attachments

    def _get_value(self, value):
        if isinstance(value, types.FunctionType):
            return value(self._context, self._receiver)
        return value

    @property
    def sender(self):
        return self._get_value(self._sender)

    @property
    def receiver_email(self):
        return self._get_value(self._receiver)

    @property
    def subject(self):
        return self._get_value(self._subject)

    @property
    def content(self):
        return self._get_value(self._content)

    @property
    def attachments(self):
        return self._get_value(self._attachments)

    @property
    def encoding(self):
        return self._get_value(self._encoding)


class Attachment(object):

    """附件"""

    @args2fields()
    def __init__(self, attachment_file, mime_type, attachment_filename=None):
        """
        :param attachment_file  作为附件的文件对象，可以是file对象或者StringIO对象，
                                如果是字符串，那么将作为文件路径进行加载
        :param mime_type  附件的mime type，比如application/octet-stream
        :param attachment_filename  附件所使用的文件名
        """
        if attachment_filename is None:
            if isinstance(attachment_file, types.StringTypes):
                self._attachment_filename = os.path.split(
                    attachment_file)[1]
            elif isinstance(attachment_file, types.FileType):
                self._attachment_filename = os.path.split(
                    attachment_file.name)[1]
            else:
                raise InvalidArgumentException(
                    u"必须制定attachement_filename参数作为附件文件名")

    def build_mime_object(self):
        """构建Mime对象"""
        mime_type = self._mime_type.split("/")
        mime = MIMEBase(mime_type[0], mime_type[1])
        mime.set_payload(self._gen_payload())
        encoders.encode_base64(mime)
        mime.add_header(
            'Content-Disposition',
            'attachment; filename="{}"'.format(self._attachment_filename))
        return mime

    def _gen_payload(self):
        # 文件类型的情况
        if isinstance(self._attachment_file, types.FileType):
            try:
                return self._attachment_file.read()
            finally:
                self._attachment_file.close()
        # 字符串路径的情况
        elif isinstance(self._attachment_file, types.StringTypes):
            with open(self._attachment_file, "r") as f:
                return f.read()
        # StringIO or cStringIO
        else:
            self._attachment_file.seek(0)
            return self._attachment_file.read()



# coding: utf-8

from girlfriend.testing import GirlFriendTestCase
from girlfriend.data.table import (
    Title,
    ListTable,
    ObjectTable,
    DictTable
)
from girlfriend.data.exception import InvalidSizeException


class ListTableTestCase(GirlFriendTestCase):

    def setUp(self):
        titles = (
            Title("id"),
            Title("name"),
            Title("age"),
            Title("grade"),
        )
        self.table = ListTable("test_table", titles)

        self.data = [
            (1, "Sam", 26, "A"),
            (2, "Jack", 31, "B"),
            (3, "James", 32, "C")
        ]

        self.wrapped_table = ListTable("test_table", titles, self.data)

    def test_append(self):
        """测试append方法
        """
        self.table.append([1, "SamChi", 26, "A"])
        self.assertEquals(
            (
                self.table[0].id,
                self.table[0].name,
                self.table[0].age,
                self.table[0].grade
            ),
            (1, "SamChi", 26, "A")
        )

        self.failUnlessException(
            InvalidSizeException,
            ListTable.append,
            self.table,
            [2, "James", 31, "X", 0])

        self.wrapped_table.append((4, "Betty", 21, "A"))
        self.assertEquals((4, "Betty", 21, "A"), self.data[-1])
        # test multi get
        self.assertEquals(
            self.wrapped_table[-1]["id", "name", "age", "grade"],
            (4, "Betty", 21, "A")
        )

    def test_iter(self):
        """测试迭代器
        """
        data = [
            (1, "Sam", 26, "A"),
            (2, "Jack", 31, "B"),
            (3, "James", 32, "C")
        ]

        for rowlist in data:
            self.table.append(rowlist)

        for idx, row in enumerate(self.table):
            self.assertEquals(
                (row.id, row.name, row.age, row.grade), data[idx])


class Student(object):

    def __init__(self, id, name, grade):
        self.id = id
        self.name = name
        self.grade = grade


class ObjectTableTestCase(GirlFriendTestCase):

    def setUp(self):
        titles = (
            Title("id", u"编号"),
            Title("name", u"姓名"),
            Title("grade", u"年级")
        )

        self.table = ObjectTable("student_table", titles)

        students = (
            Student(1, "Sam", 1),
            Student(2, "Jack", 1),
            Student(3, "Peter", 2)
        )

        self.students = students

        self.wrapped_table = ObjectTable("student_table", titles, students)

    def test_append(self):
        self.table.append(Student(4, "James", 2))
        self.assertEquals(self.table[-1].id, 4)
        self.assertEquals(self.table[0][1], "James")
        self.assertEquals(self.table[0]["grade"], 2)

    def test_iter(self):
        for idx, row in enumerate(self.wrapped_table):
            std = self.students[idx]
            self.assertEquals(
                row["id", "name", "grade"],
                (std.id, std.name, std.grade)
            )


class DictTableTestCase(GirlFriendTestCase):

    def setUp(self):
        titles = (
            Title("id", u"编号"),
            Title("name", u"姓名"),
            Title("grade", u"年级")
        )

        self.table = DictTable("student_table", titles)

        students = (
            {"id": 1, "name": "Sam", "grade": 1},
            {"id": 2, "name": "Jack", "grade": 1},
            {"id": 3, "name": "Peter", "grade": 2}
        )

        self.students = students
        self.wrapped_table = DictTable("student_table", titles, students)

    def test_append(self):
        self.table.append({"id": 4, "name": "James", "grade": 2})
        self.assertEquals(self.table[-1].id, 4)
        self.assertEquals(self.table[0][1], "James")
        self.assertEquals(self.table[0]["grade"], 2)

    def test_iter(self):
        for idx, row in enumerate(self.wrapped_table):
            record = self.students[idx]
            self.assertEquals(
                row["id", "name", "grade"],
                tuple(record[k] for k in ("id", "name", "grade"))
            )



# coding: utf-8

"""数据结构定义所需要用到的异常类
"""

from girlfriend.exception import GirlFriendSysException


class IndexOutOfBoundsException(GirlFriendSysException):
    """索引越界时抛出此异常
    """
    pass


class MissingKeyException(GirlFriendSysException):
    """在key-value的结构中，当key所引用的数据不存在时抛出此异常
    """
    pass


class InvalidSizeException(GirlFriendSysException):
    """当目标集合的尺寸不符合要求时抛出此异常
    """
    pass



# coding: utf-8

"""GirlFriend testing cases，
之所以将test放在girlfriend包下，是为了可以方便将test包也进行安装，
这样围绕girlfriend的第三方项目在开发时也可以复用这些测试工具以及fixtures。
"""

import os
import os.path
import fixtures
from termcolor import colored


class GirlFriendTestCase(fixtures.TestWithFixtures):

    def color_print(self, msg, color="red"):
        print colored(msg, color)

    def failUnlessException(self, exception_type, logic, *args, **kws):
        """运行指定的函数，如果抛出指定的异常，那么测试通过
           如果没有抛出，那么测试失败。
           :param logic 要测试的可执行对象
           :param args  可执行对象所需要的参数
           :param exception_type 期待抛出的异常
        """
        try:
            logic(*args, **kws)
        except exception_type as e:
            print
            mark = colored("[Exception message] ", "yellow", attrs=['bold'])
            print mark, colored(unicode(e), "yellow")
        else:
            self.fail(
                "Expected exception {} not happened".format(exception_type))

    def assertIsDir(self, file_path):
        """测试文件路径是否是目录类型
        """
        if not os.path.isdir(file_path):
            self.fail(u"The path {} is not a directory".format(file_path))

    def assertFileExist(self, file_path):
        """测试文件路径是否存在
        """
        if not os.path.exists(file_path):
            self.fail(u"The file {} is not existed".format(file_path))

    def assertFileAccess(self, file_path, expected_access):
        """测试文件权限
           :param expected_access 期望的文件权限，接受数字表示，比如八进制0644
        """
        stat = os.stat(file_path)
        file_access = stat.st_mode & 0777
        if file_access != expected_access:
            self.fail((
                "The expected access is {}, "
                "but the file access is {}"
            ).format(oct(expected_access), oct(file_access)))



# coding: utf-8

import xlrd
import time
import random
import types
import xlsxwriter
from StringIO import StringIO
from girlfriend.util.lang import (
    args2fields,
    SequenceCollectionType
)
from girlfriend.data.table import AbstractTable
from girlfriend.plugin.data import AbstractDataReader


class ExcelReaderPlugin(object):

    """该插件读取Excel文件，并转换为需要的数据结构
    """

    name = "read_excel"

    def execute(self, context, filepath, *sheet_readers):
        workbook = xlrd.open_workbook(filepath)
        return [reader(context, workbook) for reader in sheet_readers]


class SheetR(AbstractDataReader):

    """读取Sheet"""

    @args2fields()
    def __init__(self, sheetname, record_handler=None, record_filter=None,
                 result_wrapper=None, skip_first_row=False, variable=None):
        pass

    def __call__(self, context, workbook):
        worksheet = workbook.sheet_by_name(self._sheetname)
        result = []
        for row_index in xrange(0, worksheet.nrows):
            if self._skip_first_row and row_index == 0:
                continue
            record = [None] * worksheet.ncols
            for column_index in xrange(0, worksheet.ncols):
                value = worksheet.cell(row_index, column_index).value
                record[column_index] = value
            self._handle_record(record, result.append)
        return self._handle_result(context, result)


class ExcelWriterPlugin(object):

    """该插件输出xlsx格式文件
    """

    name = "write_excel"

    def execute(self, context, filepath, sheets=None, workbook_handler=None):
        if filepath is None:
            # 不指定则随机生成文件名
            filepath = "/tmp/{}_{}.xlsx".format(
                int(time.time()), random.randint(100, 999))
            workbook = xlsxwriter.Workbook(filepath)
        elif filepath.startswith("$"):
            output = StringIO()
            workbook = xlsxwriter.Workbook(output, {'in_memory': True})
            context[filepath[1:]] = output
        else:
            workbook = xlsxwriter.Workbook(filepath)

        for sheet in sheets:
            sheet(context, workbook)

        if workbook_handler:
            workbook_handler(workbook)

        workbook.close()
        return filepath


class SheetW(object):

    """写入Sheet"""

    @args2fields()
    def __init__(self, table=None, sheet_name=None,
                 style=None, sheet_handler=None):
        pass

    def __call__(self, context, workbook):
        sheet = self._generate_sheet(workbook)
        if self._table is not None:
            if isinstance(self._table, AbstractTable):
                self._handle_table(sheet, workbook)
            else:
                self._handle_sequence(sheet, workbook)
        if self._sheet_handler:
            self._sheet_handler(workbook, sheet)

    def _handle_table(self, sheet, workbook):
        # 先写入标题
        for idx, title in enumerate(self._table.titles):
            style = self._get_style(0, idx, workbook)
            sheet.write(0, idx, title.title, style)
        self._handle_sequence(sheet, workbook, start=1)

    def _handle_sequence(self, sheet, workbook, start=0):
        for row_index, row in enumerate(self._table, start=start):
            for column_index, column in enumerate(row):
                style = self._get_style(row_index, column_index, workbook)
                sheet.write(row_index, column_index, column, style)

    def _get_style(self, row_index, column_index, workbook):

        if not self._style:
            return None

        style = {}
        for s in self._style:
            if s.match(row_index, column_index):
                style.update(s.style_dict)
        if style:
            return workbook.add_format(style)
        else:
            return None

    def _generate_sheet(self, workbook):
        sheet_name = None
        if self._sheet_name:
            sheet_name = self._sheet_name
        else:
            sheet_name = self._table.name
        if sheet_name is None:
            return workbook.add_worksheet()
        else:
            return workbook.add_worksheet(sheet_name)


class CellStyle(object):

    """单元格样式"""

    def __init__(self, selector, style_dict):
        self._selector = selector
        self._style_dict = style_dict
        self._formatter = None

    def match(self, row_index, column_index):
        if isinstance(self._selector, types.FunctionType):
            return self._selector(row_index, column_index)
        elif isinstance(self._selector, int):
            return row_index == self._selector
        elif isinstance(self._selector, SequenceCollectionType):
            row_condition, column_condition = self._selector
            row_matched, column_matched = (
                self._match(row_index, row_condition),
                self._match(column_index, column_condition)
            )
            return row_matched and column_matched

    def _match(self, index, condition):
        if condition is None:
            return True
        if isinstance(condition, int):
            return index == condition
        elif isinstance(condition, SequenceCollectionType):
            return condition[0] <= index <= condition[-1]

    @property
    def style_dict(self):
        return self._style_dict



# coding: utf-8

from girlfriend.util.time import parse_time_unit
from girlfriend.testing import GirlFriendTestCase


class TimeUtilTestCase(GirlFriendTestCase):

    def test_parse_time_unit(self):
        self.assertEquals(parse_time_unit("2d"), 2 * 24 * 60 * 60)
        self.assertEquals(parse_time_unit("3h"), 3 * 60 * 60)
        self.assertEquals(parse_time_unit("5m"), 5 * 60)
        self.assertEquals(parse_time_unit("30000s"), 30000)



# coding: utf-8

"""参数验证相关工具
"""

import re
import ujson
import types
import numbers
from girlfriend.util.lang import args2fields
from girlfriend.exception import InvalidArgumentException


class Rule(object):

    """描述参数验证规则，并执行验证过程
    """

    @args2fields()
    def __init__(self, name,
                 type=None,
                 required=False, min=None, max=None,
                 regex=None, logic=None, default=None):
        """
        :param name      参数名称，通常用于错误提示
        :param required  如果为True，那么参数是必须的
        :param min       如果是字符串，那么该参数为最小长度(等于此长度合法)，
                         如果是数字(numbers.Number类型)，那么为该参数最小值(等于此值算合法)
        :param max       同上
        :param regex     正则验证
        :param type      类型验证，多个参数可以传递元组
        :param logic     谓词函数，满足更加复杂的业务验证需要，比如查查数据库邮箱是否存在等等
                         该谓词函数并非返回True和False，如果有错误，那么返回错误消息的字符串，
                         如果没有错误，那么直接返回None
        :param default   该项的默认值
        """
        pass

    @property
    def name(self):
        return self._name

    @property
    def default(self):
        return self._default

    @property
    def required(self):
        return self._required

    def validate(self, value):
        """执行验证
           :param value 要验证的值
        """
        if self._required and self._is_empty(value):
            raise InvalidArgumentException(
                u"参数 '{}' 的值是必须的，不能为空".format(self._name))

        # 如果非必须并且为空，那么接下来的验证就不必运行了
        if self._is_empty(value):
            return

        # 检查类型
        self._validate_type(value)

        # 检查大小、长度
        self._validate_min_max(value)

        # 检查正则
        self._validate_regex(value)

        # 检查逻辑
        self._validate_logic(value)

    def _validate_type(self, value):
        if not self._type:
            return
        if not isinstance(value, self._type):
            raise InvalidArgumentException(
                u"参数 '{name}' 的类型不正确，只允许以下类型：{types}".format(
                    name=self._name,
                    types=self._type
                )
            )

    def _validate_min_max(self, value):
        if self._min is not None:
            if isinstance(value, numbers.Number):
                if self._min > value:
                    raise InvalidArgumentException(
                        u"参数 '{name}' 的值不能小于{min}".format(
                            name=self._name, min=self._min)
                    )
            else:
                if self._min > len(value):
                    raise InvalidArgumentException(
                        u"参数 '{name}' 的长度不能小于{min}".format(
                            name=self._name, min=self._min)
                    )
        if self._max is not None:
            if isinstance(value, numbers.Number):
                if self._max < value:
                    raise InvalidArgumentException(
                        u"参数 '{name}' 的值不能大于{max}".format(
                            name=self._name, max=self._max)
                    )
            else:
                if self._max < len(value):
                    raise InvalidArgumentException(
                        u"参数 '{name}' 的长度不能大于{max}".format(
                            name=self._name, max=self._max)
                    )

    def _validate_regex(self, value):
        if not self._regex:
            return
        value = str(value)
        if not re.search(self._regex, value):
            raise InvalidArgumentException(
                u"参数 '{name}' 不符合正则表达式'{regex}'".format(
                    name=self._name, regex=self._regex)
            )

    def _validate_logic(self, value):
        if self._logic is None:
            return
        msg = self._logic(value)
        if msg:
            raise InvalidArgumentException(msg)

    def _is_empty(self, value):
        """判断一个值是否为空
           如果值为None，那么返回True
           如果值为空字符串，那么返回True
           如果值为0, 那么不算空，返回False
        """
        if value is None:
            return True
        if isinstance(value, types.StringType) and not value:
            return True
        return False


def be_json(name):
    def _be_json(value):
        try:
            ujson.loads(value)
        except:
            return u"参数 '{}' 必须是json格式".format(name)
    return _be_json



# coding: utf-8

"""json读写插件
"""

import re
import types
import ujson
import requests
from girlfriend.util.lang import args2fields
from girlfriend.util.resource import HTTP_SCHEMA
from girlfriend.plugin.data import (
    AbstractDataReader,
    AbstractDataWriter
)
from girlfriend.exception import InvalidArgumentException


class JSONReaderPlugin(object):

    """可以从文件或者Web URL中加载json对象，并进行格式转换
       支持常见的json文件格式
    """

    name = "read_json"

    def execute(self, context, *json_readers):
        return [reader(context) for reader in json_readers]


JSON_REGEX = re.compile(r"\{.*?\}")


class JSONR(AbstractDataReader):

    @args2fields()
    def __init__(self, path, style,
                 record_handler=None, record_filter=None, result_wrapper=None,
                 variable=None):
        """
        :param  context 上下文对象
        :param  path  加载路径，可以是文件路径，也可以是web url
        :param  style  json数据格式，允许三种格式：
                        1. line: 文件每行是一个json对象
                        2. array: 文件内容是一个json数组
                        3. extract:property 文件是一个json对象，但是只提取某一部分进行处理
                        4. block: 区块，不在同一行
        :param  record_handler  行处理器，返回的每行都是字典对象，通过该函数可以进行包装
                                如果返回None，那么将对该行忽略
        :param  record_filter   行过滤器
        :param  result_wrapper 对最终结果进行包装
        :param  variable  结果写入上下文的变量名，如果为None，那么将返回值交给框架自身来保存
        """
        pass

    def __call__(self, context):
        result = []

        # 基于文件的逐行加载
        if self._style == "line" and not self._path.startswith(HTTP_SCHEMA):
            with open(self._path, "r") as f:
                for line in f:
                    line = line.strip()
                    if not line or line.startswith(("#", "//", ";")):
                        continue
                    record = ujson.loads(line)
                    self._handle_record(record, result.append)
        else:
            json_content = None
            # 从不同的来源加载json对象
            if self._path.startswith(HTTP_SCHEMA):
                json_content = requests.get(self._path).text
            else:
                with open(self._path, "r") as f:
                    json_content = f.read()

            json_content = json_content.strip()

            # 按行读取
            if self._style == "line":
                for line in json_content.splitlines():
                    line = line.strip()
                    if not line:
                        continue
                    record = ujson.loads(line)
                    self._handle_record(record, result.append)
            # 按块读取
            if self._style == "block":
                json_buffer = []
                in_block = False
                for char in json_content:
                    if char == "{":
                        in_block = True
                        json_buffer.append(char)
                    elif char == "}" and in_block:
                        json_buffer.append(char)
                        try:
                            record = ujson.loads("".join(json_buffer))
                        except ValueError:
                            continue
                        else:
                            self._handle_record(record, result.append)
                            json_buffer = []
                            in_block = False
                    elif in_block:
                        json_buffer.append(char)
            # 按数组读取
            elif self._style == "array":
                json_array = ujson.loads(json_content)
                for record in json_array:
                    self._handle_record(record, result.append)
            # 使用属性提取器
            elif self._style.startswith("extract:"):
                json_obj = ujson.loads(json_content)
                keys = self._style[len("extract:"):].split(".")
                for key in keys:
                    json_obj = json_obj[key]
                for record in json_obj:
                    self._handle_record(record, result.append)

        return self._handle_result(context, result)


class JSONWriterPlugin(object):

    name = "write_json"

    def execute(self, context, *json_writers):
        for json_writer in json_writers:
            json_writer(context)


class JSONW(AbstractDataWriter):

    @args2fields()
    def __init__(self, path, style, object,
                 record_handler=None, record_filter=None,
                 http_method="post", http_field=None, variable=None):
        """
        :param path 写入路径，默认为文件路径，如果是HTTP或者HTTPS开头，那么将会POST到对应的地址
        :param style 写入格式，line - 按行写入 array - 作为json数组写入 object - 作为单独对象写入
        :param table 要操作的对象，可以是具体的对象，也可以是context中的变量名
        :param record_handler 行处理器，可以在此进行格式转换，比如把时间对象转换为字符串
        :param record_filter 行过滤器
        :param http_method http写入方法，默认为POST，可以指定PUT
        :param variable 将json写入上下文变量
        """
        pass

    def __call__(self, context):
        if (self._style == "line" and self._path and
                not self._path.startswith(HTTP_SCHEMA)):
            with open(self._path, "w") as f:
                for row in self._object:
                    row = self._handle_record(row)
                    f.write(ujson.dumps(row) + "\n")
            return

        # json文本
        json_text = ""

        if isinstance(self._object, types.FunctionType):
            self._object = self._object(context)

        if self._style == "object":
            json_text = ujson.dumps(self._object)

        result = []
        for row in self._object:
            row = self._handle_record(row, result.append)

        # 数组格式直接dump
        if self._style == "array":
            json_text = ujson.dumps(result)

        # line格式
        if self._style == "line":
            string_buffer = []
            for row in self._object:
                row = self._handle_record(row)
                string_buffer.append(ujson.dumps(row))
            json_text = "\n".join(string_buffer)

        if self._path is None:
            if self._variable:
                context[self._variable] = json_text
                return
            else:
                raise InvalidArgumentException(
                    u"当path为None时，必须指定一个有效的variable")

        if self._path.startswith(HTTP_SCHEMA):
            if self._http_method.lower() == "post":
                if self._http_field:
                    requests.post(
                        self._path, data={self._http_field: json_text})
                elif self._style == "line":
                    requests.post(self._path, data=json_text)
                else:
                    requests.post(self._path, json=json_text)
            elif self._http_method.lower() == "put":
                requests.put(self._path, json=json_text)
        else:
            with open(self._path, "w") as f:
                f.write(json_text)

        if self._variable:
            context[self._variable] = json_text



# coding: utf-8

"""
该工具用于快速生成配置文件和用户工作目录
"""

import sys
import os.path
import argparse
from girlfriend import VERSION
from girlfriend.exception import GirlFriendBizException
from girlfriend.util import config, script
from girlfriend.util.file_template import Dir, File
from girlfriend.tools.ascii_art import MM_LINGMUAILI
from girlfriend.tools._template import TEMPLATE_TEMPLATE

WELCOME = """
{mm}

=== Girl Friend ===

版本：{version}
Bug或者好的想法欢迎提交GitHub issue
或者发送邮件到 hongze.chi@gmail.com
谢谢 :)
""".format(mm=MM_LINGMUAILI, version=VERSION)

HOME_WORKSPACE = os.path.join(os.environ["HOME"], ".gf")


def main():
    cmd_args = parse_cmd_args()
    try:
        if cmd_args.f:
            gen_config_file(cmd_args.f)
            script.show_msg_and_exit(
                u"已经成功创建配置文件'{}'".format(cmd_args.f), "green")
        else:
            gen_home_workspace()
            print WELCOME
            script.show_msg_and_exit(
                u"默认工作目录以及默认配置文件创建成功，请查看 '{}'".format(HOME_WORKSPACE),
                "green")
    except GirlFriendBizException as biz_e:
        script.show_msg_and_exit(unicode(biz_e), "yellow")
    except Exception as sys_e:
        script.show_traceback_and_exit(unicode(sys_e))


def parse_cmd_args():
    parser = argparse.ArgumentParser(description=__doc__.decode("utf-8"))
    parser.add_argument("-f", default="", help=u"生成配置文件的路径")
    ns = parser.parse_args(sys.argv[1:])
    return ns


def gen_config_file(file_path):
    """依据模板生成配置文件
       :param file_path 配置文件路径
    """
    path, filename = os.path.split(file_path)
    File(filename, content=config.CONFIG_EXAMPLE).makeme(path)


def gen_home_workspace():
    Dir(".gf", elements=(
        Dir("project_templates", elements=(
            Dir("mine"),
            Dir("company"),
            File("template.py", content=TEMPLATE_TEMPLATE),
        )),  # 用于保存常用的工程模板
        File("gf.cfg", content=config.CONFIG_EXAMPLE)
    )).makeme(os.environ["HOME"])

if __name__ == "__main__":
    main()



# coding: utf-8

"""该模块中描述了插件的基本数据结构和生命周期，
并提供插件编写和管理的基本工具。
"""

import types
import functools
from girlfriend.exception import (
    GirlFriendBizException,
    GirlFriendSysException
)
from girlfriend.util.lang import SequenceCollectionType
from stevedore import extension


class Plugin(object):

    """插件

       GirlFriend中的插件的生命周期

           sys_prepare 在整个系统中只执行一次,用于做一些初始化工作
           execute 用于执行具体的插件逻辑,并通过上下文来进行参数传递
           sys_cleanup 在整个系统中也只执行一次,用于做一些清理工作

       GirlFriend插件的公共属性

           name 插件名称,要确保插件名称的独一无二,可以使用倒置域名法命名
           status 插件状态
           description 插件描述,用于输出帮助信息
    """

    @classmethod
    def wrap_function(cls, name, description,
                      execute, sys_prepare=None, sys_cleanup=None,
                      args_validator=None, config_validator=None):
        """将函数对象包装成插件对象
        :param name: 插件名称
        :param description: 插件描述
        :param execute: 执行函数
        :param sys_prepare: sys_prepare函数,可选
        :param sys_cleanup: sys_cleanup函数,可选
        :param args_validator: 执行参数验证器
        :param config_validator: 配置验证器
        :return: 包装后的Plugin对象
        """
        execute = Plugin.__check_function(
            "execute",
            execute,
            lambda arg_count: arg_count >= 1,
            u"execute函数至少要有一个参数用来接受上下文对象")

        if sys_prepare:
            sys_prepare = Plugin.__check_function(
                "sys_prepare",
                sys_prepare,
                lambda arg_count: arg_count == 1,
                u"sys_prepare函数只能有且只有一个参数来接受配置信息"
            )

        if sys_cleanup:
            sys_cleanup = Plugin.__check_function(
                "sys_cleanup",
                sys_cleanup,
                lambda arg_count: arg_count == 1,
                u"sys_cleanup函数只能有且只有一个参数来接受配置信息"
            )

        return cls(
            name=name,
            description=description,
            execute=execute,
            sys_prepare=sys_prepare,
            sys_cleanup=sys_cleanup,
            args_validator=args_validator,
            config_validator=config_validator
        )

    @staticmethod
    def __check_function(event_name, function, check_arg, msg):
        """检查传入的函数是否符合指定格式
        """
        if not isinstance(function, types.FunctionType):
            raise InvalidPluginException(u"'{}' 参数必须是函数类型".format(event_name))
        if not check_arg(function.func_code.co_argcount):
            raise InvalidPluginException(msg)
        return function

    @classmethod
    def wrap_class(cls, clazz, *args, **kws):
        """将类对象包装为插件对象,例如:

            class TestPlugin(object):

                name = "testing"

                '''
                This is a testing plugin
                '''

                def __init__(self, a, b):
                    self.a = a
                    self.b = b

                def sys_prepare(self, config):
                    print "sys_prepare method"

                def execute(self, context, *args, **kws):
                    print "execute method"

                def sys_cleanup(self, config):
                    print "sys_cleanup method"

            plugin = Plugin.wrap_class(TestPlugin, a=1, b=2)

            Plugin.wrap_class会使用给予的参数来调用插件类的构造函数,
            创建出该类的对象,并把生命周期方法的调用委托给这个新创建的对象,
            name属性默认会取类名,也可以跟上面一样指定一个name类属性,
            description属性会取类的__doc__属性
            其中,execute方法是必须的,sys_prepare和sys_cleanup两个方法是可选的


        :param clazz: 要包装的类对象
        :param args: 构造函数的参数列表
        :param kws: 构造函数的参数列表
        :return: 包装后的Plugin对象
        """

        class_name = clazz.__name__

        if not hasattr(clazz, "execute"):
            raise InvalidPluginException(
                u"类 '{}' 缺失execute方法,无法构建插件".format(class_name))

        # 构建实例
        instance = clazz(*args, **kws)

        # 检查并包装方法
        execute = Plugin.__check_and_wrap_method(
            instance,
            getattr(clazz, "execute", None),
            lambda c: c >= 1,
            u"{}.execute 除了self至少要包含一个参数用来传递context对象".format(class_name)
        )

        sys_prepare = Plugin.__check_and_wrap_method(
            instance,
            getattr(clazz, "sys_prepare", None),
            lambda c: c == 1,
            u"{}.sys_prepare 除了self只能有且只有一个参数来接受config对象".format(class_name)
        )

        sys_cleanup = Plugin.__check_and_wrap_method(
            instance,
            getattr(clazz, "sys_cleanup", None),
            lambda c: c == 1,
            u"{}.sys_cleanup 除了self只能有且只有一个参数来接受config对象".format(class_name)
        )

        # 解决其余属性
        plugin_name = class_name
        if hasattr(clazz, "name"):
            name = getattr(clazz, "name")
            if not name or not name.strip():
                raise InvalidPluginException(
                    u"{}.name 属性不能为空".format(class_name))
            plugin_name = name.strip()

        args_validator = getattr(clazz, "args_validator", tuple())
        config_validator = getattr(clazz, "config_validator", tuple())

        # 集齐了各路神器,召唤神龙!
        return cls(
            name=plugin_name,
            description=clazz.__doc__,
            execute=execute,
            sys_prepare=sys_prepare,
            sys_cleanup=sys_cleanup,
            args_validator=args_validator,
            config_validator=config_validator
        )

    @staticmethod
    def __check_and_wrap_method(instance, mtd, check_arg, msg):
        if not mtd:
            return None
        if not check_arg(mtd.im_func.func_code.co_argcount - 1):
            raise InvalidPluginException(msg)
        return functools.partial(mtd, instance)

    @classmethod
    def wrap_module(cls, module):
        """包装模块对象为Plugin对象,例如test_plugin.py:

            '''测试模块'''

            name = "test_plugin"

            def sys_prepare(config):
                pass

            def execute(config):
                pass

            def sys_cleanup(config):
                pass


        module对象必须包含一个execute函数
        sys_prepare跟sys_cleanup是可选的

        会默认将模块的名称作为插件名称,如果模块中包含了name属性,
        那么name的值将作为插件的名称

        模块的__doc__属性将作为插件的description属性

        :param module: 要包装的模块对象
        :return: 包装后的Plugin对象
        """

        module_name = module.__name__

        if not hasattr(module, "execute"):
            raise InvalidPluginException(
                u"模块 '{}' 必须包含一个execute函数".format(module_name))

        # 检查模块中各个函数是否满足要求

        execute = Plugin.__check_function(
            "execute",
            module.execute,
            lambda c: c >= 1,
            u"模块 '{}' 的execute函数至少要有一个参数用于接收context对象".format(module_name)
        )

        sys_prepare, sys_cleanup = None, None

        if hasattr(module, "sys_prepare"):
            sys_prepare = Plugin.__check_function(
                "sys_prepare",
                module.sys_prepare,
                lambda c: c == 1,
                u"模块 '{}' 的sys_prepare函数必须有一个参数用于接收config对象".format(
                    module_name)
            )

        if hasattr(module, "sys_cleanup"):
            sys_cleanup = Plugin.__check_function(
                "sys_cleanup",
                module.sys_cleanup,
                lambda c: c == 1,
                u"模块 '{}' 的sys_cleanup函数必须有".format(module_name)
            )

        # 获取插件名称
        plugin_name = module_name
        if hasattr(module, "name"):
            name = module.name
            if not name or not name.strip():
                raise InvalidPluginException(
                    u"模块 '{}' 的name属性不能为空".format(name))
            plugin_name = name.strip()

        args_validator = getattr(module, "args_validator", tuple())
        config_validator = getattr(module, "config_validator", tuple())

        return cls(
            name=plugin_name,
            description=module.__doc__,
            execute=execute,
            sys_prepare=sys_prepare,
            sys_cleanup=sys_cleanup,
            args_validator=args_validator,
            config_validator=config_validator
        )

    STATUS_UNPREPARED = 0  # 尚未进行初始化
    STATUS_PREPARED = 1  # 已经进行了初始化,处于可执行状态
    STATUS_DEAD = 2  # 已经进行了清理,无法再使用

    def __init__(self, name, description, execute, sys_prepare, sys_cleanup,
                 args_validator, config_validator):
        """
        :param name: 插件名称,在整个系统中,插件需要有一个独一无二的名称
        :param execute: 插件的执行逻辑,
                        接受一个context参数和其它请求参数的可执行对象
        :param sys_prepare: 初始化钩子,接受一个config参数的可执行对象
        :param sys_cleanup: 清理钩子,接受一个config参数的可执行对象
        :param args_validator: 参数验证器，接受一个rule列表，或者一个自定义的验证函数
        :param config_validator: 配置验证器
        """
        self._name = name
        self._description = description
        self._sys_prepare = sys_prepare
        self._sys_cleanup = sys_cleanup
        self._execute = execute

        # 启用默认的参数验证器
        if args_validator is None or isinstance(
                args_validator, SequenceCollectionType):
            self._args_validator = DefaultArgsValidator(args_validator)
        else:
            self._args_validator = args_validator

        # 启用默认的配置验证器
        if config_validator is None or isinstance(
                config_validator, SequenceCollectionType):
            self._config_validator = DefaultConfigValidator(config_validator)
        else:
            self._config_validator = config_validator

        self._status = Plugin.STATUS_UNPREPARED  # 尚未初始化

    @property
    def name(self):
        return self._name

    @property
    def status(self):
        return self._status

    @property
    def description(self):
        return self._description

    def sys_prepare(self, config):
        """系统初始化
        :param config: 配置信息
        """
        if self.status != Plugin.STATUS_UNPREPARED:
            raise PluginAlreadyPreparedException(
                u"插件 '{}' 已经被初始化了".format(self.name))

        # 执行配置验证
        self._config_validator(config)
        value = None
        if self._sys_prepare:
            value = self._sys_prepare(config)
        self._status = Plugin.STATUS_PREPARED
        return value

    def sys_cleanup(self, config):
        """清理工作,在系统生命周期结束时进行
        :param config: 配置信息
        """
        if self.status == Plugin.STATUS_UNPREPARED:
            raise PluginUnPreparedException(
                u"插件 '{}' 尚未初始化,不能进行清理工作".format(self.name))
        if self.status == Plugin.STATUS_DEAD:
            return
        value = None
        if self._sys_cleanup:
            value = self._sys_cleanup(config)
        self._status = Plugin.STATUS_DEAD
        return value

    def execute(self, context, *args, **kws):
        """执行逻辑
        :param context: 上下文对象,用于获取参数或者传递结果
        :param args: 参数列表
        :param kws: keyword参数列表
        """
        if self.status == Plugin.STATUS_UNPREPARED:
            raise PluginUnPreparedException(
                u"插件 '{}' 尚未初始化,不能执行".format(self.name))
        if self.status == Plugin.STATUS_DEAD:
            raise PluginAlreadyDeadException(
                u"插件 '{}' 已经被清理过了,不能执行".format(self.name))
        # 执行参数验证
        self._args_validator(*args, **kws)
        return self._execute(context, *args, **kws)

    def __repr__(self):
        return None


class DefaultArgsValidator(object):

    """默认的参数验证器
    """

    def __init__(self, rules):
        self._rules = rules
        if rules:
            self._rule_dict = {rule.name: rule for rule in rules}

    def __call__(self, *args, **kws):
        if not self._rules:
            return
        for rule, arg_value in zip(self._rules, args):
            rule.validate(arg_value)
        for arg_name, arg_value in kws.items():
            rule = self._rule_dict[arg_name]
            rule.validate(arg_value)


class DefaultConfigValidator(object):

    """默认的配置项验证器
    """

    def __init__(self, rules):
        self._rules = rules

    def __call__(self, config):
        if not self._rules or not config:
            return
        for rule in self._rules:
            section, item_name = rule.name.split(".")
            item_value = None
            if section in config:
                items = config[section]
                item_value = items[item_name]
            rule.validate(item_value)
            if item_value is None and not rule.required:
                config[section][item_name] = rule.default


class InvalidPluginException(GirlFriendSysException):

    """构建插件对象出错时抛出此异常
    """
    pass


class PluginUnPreparedException(GirlFriendSysException):

    """当插件未进行初始化时抛出此异常
    """
    pass


class PluginAlreadyPreparedException(GirlFriendSysException):

    """当插件已经被初始化时抛出此异常
    """
    pass


class PluginAlreadyDeadException(GirlFriendSysException):

    """当插件已经被清理时抛出此异常
    """
    pass


class PluginManager(object):

    """插件管理器,主要包含的功能:

            注册插件
            获取插件
            批量对插件进行初始化
            批量对插件进行清理
            替换插件

       plugin模块会提供一个默认的PluginManager实例
       并会在该实例上默认注册所有的内置插件和基于girlfriend.plugin这个namespace
       的第三方插件

       开发者可以自行创建自己的PluginManager实例,并自行约定插件规则,
       通过PluginManagerChain可以将多个PluginManager按照优先级拼接在一起
    """

    def __init__(self):
        self._plugins = {}

    def plugin(self, plugin_name):
        """获取指定的插件
        :param plugin_name: 插件名称
        :return: 插件对象
        """
        return self.__getitem__(plugin_name)

    def __getitem__(self, plugin_name):
        if plugin_name not in self._plugins:
            raise PluginNotFoundException(plugin_name)
        return self._plugins[plugin_name]

    def register(self, plugin):
        """将插件注册到容器中,相同名字的插件不允许被重复注册
           如果要替换插件,那么需要先将插件从容器中移除,然后再进行注册
        :param plugin: 要注册的插件对象
        """
        if plugin.name in self._plugins:
            raise PluginAlreadyRegisteredException(plugin.name)
        self._plugins[plugin.name] = plugin

    def sys_prepare(self, config, *plugin_names):
        """对所有插件进行初始化
        :param config: 配置信息
        :param plugin_names: 要初始化的插件名称,如果为空,那么默认初始化所有插件
        """
        if not plugin_names:
            plugin_names = self._plugins.keys()
        for plugin_name in plugin_names:
            self._plugins[plugin_name].sys_prepare(config)

    def sys_cleanup(self, config, *plugin_names):
        """对所有插件进行清理
        :param config: 配置信息
        :param plugin_names: 要清理的插件名称,如果为空,那么默认对所有插件进行清理操作
        """
        if not plugin_names:
            plugin_names = self._plugins.keys()
        for plugin_name in plugin_names:
            self._plugins[plugin_name].sys_cleanup(config)

    def remove(self, plugin_name):
        self.__delitem__(plugin_name)

    def __delitem__(self, plugin_name):
        """删除插件
        """
        if plugin_name not in self._plugins:
            raise PluginNotFoundException(plugin_name)
        del self._plugins[plugin_name]

    def replace(self, plugin):
        """如果插件不存在,那么对插件进行注册,如果插件已经存在
           那么删除旧有的插件,使用新的插件对象进行注册
        :param plugin: 插件对象
        """
        try:
            self.register(plugin)
        except PluginAlreadyRegisteredException:
            self.remove(plugin.name)
            self.register(plugin)


class PluginAlreadyRegisteredException(GirlFriendBizException):

    """向插件管理器注册插件,如果插件名称已经存在,那么抛出此异常
    """

    def __init__(self, plugin_name):
        GirlFriendBizException.__init__(
            self,
            u"插件 '{}' 已经被注册了".format(plugin_name))


class PluginNotFoundException(GirlFriendBizException):

    """从插件管理器中获取插件对象,如果找不到,那么抛出此异常
    """

    def __init__(self, plugin_name):
        GirlFriendBizException.__init__(
            self,
            u"插件 '{}' 不存在".format(plugin_name))


plugin_manager = PluginManager()


def register_entry_points_plugins(entry_point="girlfriend.plugin"):
    """将基于entry_point的第三方插件注册到默认管理器
       :param entry_point: 加载插件的entry_point
    """

    global plugin_manager

    third_party_plugin_mgr = extension.ExtensionManager(
        namespace=entry_point, invoke_on_load=False)

    for ext in third_party_plugin_mgr:
        plugin_object, plugin = ext.plugin, None

        # 插件对象类型
        if isinstance(plugin_object, Plugin):
            plugin = plugin_object

        # 模块类型
        elif isinstance(plugin_object, types.ModuleType):
            plugin = Plugin.wrap_module(plugin_object)

        # 函数类型
        elif isinstance(plugin_object, types.FunctionType):
            plugin = Plugin.wrap_function(
                ext.name, plugin_object.__doc__, plugin_object)

        # 类类型
        elif isinstance(plugin_object, (types.ClassType, types.TypeType)):
            plugin = Plugin.wrap_class(plugin_object)

        # 不被支持的插件类型
        else:
            raise InvalidPluginException(
                u"插件 '{}' 的类型 '{}'不被支持".format(
                    ext.name,
                    type(plugin_object).__name__)
            )

        plugin_manager.register(plugin)


def auto_register():
    """自动注册内置的插件和基于entry_point的第三方插件
    """
    register_entry_points_plugins()


# 默认自动注册
auto_register()


class PluginManagerChain(object):

    """插件管理器链,当需要从多个插件管理器中获取插件时
       可将多个管理器按照优先级的先后传递给该对象,然后就可以按照优先级获取插件了
       总之,行为类似于Python3.x中的collections.ChainMap
    """

    def __init__(self, *plugin_managers):
        """
        :param plugin_managers: 插件管理器列表, 优先级高的管理器在前
        """
        self._plugin_managers = plugin_managers

    def plugin(self, plugin_name):
        """从插件管理器链中获取插件对象
        :param plugin_name: 插件名称
        :return 优先级最高的插件对象
        """
        return self.__getitem__(plugin_name)

    def __getitem__(self, plugin_name):
        for pm in self._plugin_managers:
            try:
                return pm.plugin(plugin_name)
            except PluginNotFoundException:
                continue
        raise PluginNotFoundException(plugin_name)

    def sys_prepare(self, config, *plugin_names):
        """将chain中所有的PluginManager依次全部初始化
        :param config: 配置对象
        :param plugin_names: 要初始化的插件名称
        """
        if plugin_names:
            for plugin_name in plugin_names:
                self.plugin(plugin_name).sys_prepare(config)
        else:
            for plugin_mgr in self._plugin_managers:
                plugin_mgr.sys_prepare(config)

    def sys_cleanup(self, config, *plugin_names):
        """将chain中所有的PluginManager依次全部清理
        :param config: 配置对象
        :param plugin_names: 要初始化的插件名称
        """
        if plugin_names:
            for plugin_name in plugin_names:
                self.plugin(plugin_name).sys_cleanup(config)
        else:
            for plugin_mgr in self._plugin_managers:
                plugin_mgr.sys_cleanup(config)



# coding: utf-8

"""数据转换插件抽象
"""

from abc import (
    ABCMeta,
    abstractmethod
)


class AbstractDataHandler(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __call__(self, context):
        pass

    def _handle_record(self, record, collector=None):
        """处理记录
        :param record 记录
        :param collector 收集器，回调函数或者方法，比如list.append，set.add
        """
        # 执行过滤器
        if self._record_filter is not None:
            if not self._record_filter(record):
                return

        # 对行进行数据转换
        if self._record_handler is not None:
            record = self._record_handler(record)

        if collector is not None:
            collector(record)

        return record


class AbstractDataReader(AbstractDataHandler):

    __metaclass__ = ABCMeta

    def _handle_result(self, context, result):
        """对最终结果进行包装
        """
        if self._result_wrapper is not None:
            result = self._result_wrapper(result)
        if self._variable:
            context[self._variable] = result
        return result


class AbstractDataWriter(AbstractDataHandler):

    __metaclass__ = ABCMeta

    pass



#! /usr/bin/env python
# coding: utf-8

import setuptools

install_requires = [
    "SQLAlchemy >= 1.0.0",
    "prettytable",
    "httpretty",
    "ujson",
    "termcolor >= 1.1.0",
    "requests >= 2.7.0",
    "fixtures >= 1.4.0",
    "stevedore >= 1.7.0",
    "python_daemon",
    "XlsxWriter >= 0.8.4",
    "xlrd >= 0.9.3",
    "argparse >= 1.3.0",
]

setuptools.setup(
    name="girlfriend",
    version="0.0.4",
    author="Sam Chi",
    author_email="chihongze@gmail.com",
    description=(
        "A pure Python girlfriend "
        "she can help you build operation scripts, "
        "send data report, monitor the system "
        "and do a lot of things you undreamed!"
        "The most important, "
        "her heart(core lib) is all completely free!"
    ),
    license="MIT",
    packages=setuptools.find_packages("."),
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "gf_workflow = girlfriend.tools.gf_workflow:main",
            "gf_config = girlfriend.tools.gf_config:main",
            "gf_gen = girlfriend.tools.gf_gen:main"
        ],

        # builtin plugins
        "girlfriend.plugin": [
            # db plugin
            "orm_query = girlfriend.plugin.orm:OrmQueryPlugin",

            # table plugin
            "table_adapter = girlfriend.plugin.table:TableAdapterPlugin",
            "column2title = girlfriend.plugin.table:TableColumn2TitlePlugin",
            "print_table = girlfriend.plugin.table:PrintTablePlugin",
            "concat_table = girlfriend.plugin.table:ConcatTablePlugin",
            "join_table = girlfriend.plugin.table:JoinTablePlugin",
            "split_table = girlfriend.plugin.table:SplitTablePlugin",
            "html_table = girlfriend.plugin.table:HTMLTablePlugin",

            # json plugin
            "read_json = girlfriend.plugin.json:JSONReaderPlugin",
            "write_json = girlfriend.plugin.json:JSONWriterPlugin",

            # excel plugin
            "read_excel = girlfriend.plugin.excel:ExcelReaderPlugin",
            "write_excel = girlfriend.plugin.excel:ExcelWriterPlugin",

            # csv plugin
            "read_csv = girlfriend.plugin.csv:CSVReaderPlugin",
            "write_csv = girlfriend.plugin.csv:CSVWriterPlugin",

            # email plugin
            "send_mail = girlfriend.plugin.mail:SendMailPlugin",
        ],

        # builtin workflow
        "girlfriend.workflow": [
            "sqlreport = girlfriend.workflow.builtin:sqlreport",
        ]
    },
)



# coding: utf-8

"""封装常用的设计模式以及对内置函数和方法进行约定性的简化
"""

import types
import inspect
from functools import wraps
from girlfriend.exception import GirlFriendSysException

_singletons = {}


def singleton(clazz):
    """单例修饰器，被修饰的类在系统中都是单例的
       非线程安全，请勿用在多线程环境当中
    """

    @wraps(clazz)
    def constructor(*args, **kws):
        global _singletons
        instance = _singletons.get(clazz)
        if instance is None:
            instance = clazz(*args, **kws)
            _singletons[clazz] = instance
        return instance

    return constructor


class DelegateMeta(type):

    """该元类用于实现委托
       基本用法:

       class A(object):

            __metaclass__ = DelegateMeta

            def __init__(self, delegate):
                self.delegate = delegate

        lst = [1,2,3]
        a = A(lst)
        a.append(4)

        对象a的append方法会自动委托到lst对象的append方法。
        这样可以满足多数情况，但是碰到内置方法比如__getitem__无法自动实现委托
        如果要委托内置方法，那么需要通过类属性delegate_internal_methods去指明

        class B(object):
            __metaclass__ = DelegateMeta
            delegate_internal_methods = (
                "__getitem__",
                "__hash__",
                "__eq__"
            )

            def __init__(self, delegate):
                self.delegate = delegate

        需要值得注意的是，不要委托特殊方法：__init__、__new__
        另外还有__getattr__以及__getattribute__也不可以委托，因为DelegateMeta会用到这两个方法
        如果需要在委托类中对访问属性做控制，那么可以使用__myattr__(self, fieldname)
        对于未定义属性，DelegateMeta会优先拦截__myattr__，__myattr__通过抛出UnknownAttrError通知
        委托类进行接下来的处理。

        还可以使用delegate_methods属性显式指定委托方法:

        class C(object):
            __metaclass__ = DelegateMeta
            delegate_methods = (
                "append",
                "__getitem__",
                "__eq__"
            )

            def __init__(self, delegate):
                self.delegate = delegate

    """

    class UnknownAttrError(GirlFriendSysException):
        pass

    def __new__(cls, name, bases, attrs):
        delegate_methods = attrs.get("delegate_methods", tuple())
        if delegate_methods:
            DelegateMeta.register_delegates(delegate_methods, attrs)
            return type(name, bases, attrs)

        def getter(self, method_name):
            if "__myattr__" in attrs:
                try:
                    return self.__myattr__(method_name)
                except DelegateMeta.UnknownAttrError:
                    pass

            def method(*args, **kws):
                mtd = getattr(self.delegate, method_name)
                if mtd:
                    return mtd(*args, **kws)
                else:
                    raise AttributeError(
                        "No method found %s" % method_name)
            return method

        attrs["__getattr__"] = getter

        # 痛!
        delegate_internal_methods = attrs.get(
            "delegate_internal_methods", tuple())
        DelegateMeta.register_delegates(delegate_internal_methods, attrs)
        return type(name, bases, attrs)

    @staticmethod
    def register_delegates(delegate_methods, attrs):
        for mtd_name in delegate_methods:
            def make_method(method_name):
                def method(self, *args, **kws):
                    return getattr(self.delegate,
                                   method_name)(*args, **kws)
                return method
            attrs[mtd_name] = make_method(mtd_name)


def args2fields(private=True):
    """专门是应用于构造函数的修饰器
       可以将构造函数除self以外的参数悉数赋值给类属性

       比如

       class A(object):

            def __init__(self, a, b, c):
                self._a = a
                self._b = b
                self._c = c
                self.sum = self._a + self._b + self._c

        只要写成这样就好:

        class A(object):

            @args2fields()
            def __init__(self, a, b, c):
                self.sum = self._a + self._b + self._c

        不必再去写上面那些无聊的赋值语句了

        :param private 是否转变为私有字段，如果为True，那么会在所有字段名前加个下划线
    """

    def _field_name(arg_name):
        return "_" + arg_name if private else arg_name

    def _args2fields(constructor):

        @wraps(constructor)
        def _wrapped_constuctor(self, *args, **kws):
            args_spec = inspect.getargspec(constructor)
            for idx, arg in enumerate(args, start=1):
                arg_name = args_spec.args[idx]
                field_name = _field_name(arg_name)
                setattr(self, field_name, arg)
            for arg_name, arg in kws.items():
                field_name = _field_name(arg_name)
                setattr(self, field_name, arg)

            # 处理没有赋值的默认参数
            default_args = get_default_args(args_spec)
            if default_args:
                for arg_name, default_value in default_args.items():
                    if arg_name == "self":
                        continue
                    field_name = _field_name(arg_name)
                    if hasattr(self, field_name):
                        continue
                    setattr(self, field_name, default_value)

            constructor(self, *args, **kws)

        return _wrapped_constuctor

    return _args2fields


def get_default_args(o):
    """获取函数的默认参数名-值映射
    """
    argspec = o
    if not isinstance(o, inspect.ArgSpec):
        argspec = inspect.getargspec(o)
    if not argspec.defaults:
        return {}
    return dict(zip(argspec.args[-len(argspec.defaults):],
                    argspec.defaults))

# 线性集合类型
SequenceCollectionType = (types.ListType, types.TupleType)


def parse_context_var(context, variable_name):
    """解析上下文中的变量
       如果以'$'字符开头，那么返回上下文中的对应变量
       其它的情况会直接返回字符串
       开头两个$$连续为转义，比如'$$aa$$a'为'$aa$$a'
       :param context 上下文
       :param variable_name
    """
    if not isinstance(variable_name, str):
        return variable_name
    elif variable_name.startswith("$$"):
        return variable_name.replace("$$", "$")
    elif variable_name.startswith("$"):
        return context[variable_name[1:]]
    else:
        return variable_name
